@@ -15,7 +15,6 @@
 #include <sys/vfs.h>
 
 #include <sys/sendfile.h>
-#include <sys/mman.h>
 
 #include <sched.h>
 #include <sys/resource.h>
@@ -32,7 +31,7 @@
 
 #include "asm/types.h"
 #include "list.h"
-#include "fdset.h"
+#include "imgset.h"
 #include "file-ids.h"
 #include "kcmp-ids.h"
 #include "compiler.h"
@@ -42,7 +41,6 @@
 #include "syscall.h"
 #include "ptrace.h"
 #include "util.h"
-#include "sockets.h"
 #include "namespaces.h"
 #include "image.h"
 #include "proc_parse.h"
@@ -59,12 +57,12 @@
 #include "sk-packet.h"
 #include "cpu.h"
 #include "elf.h"
+#include "cgroup.h"
 #include "file-lock.h"
 #include "page-xfer.h"
 #include "kerndat.h"
 #include "stats.h"
 #include "mem.h"
-#include "vdso.h"
 #include "page-pipe.h"
 #include "posix-timer.h"
 #include "vdso.h"
@@ -72,6 +70,8 @@
 #include "cr-service.h"
 #include "plugin.h"
 #include "irmap.h"
+#include "sysfs_parse.h"
+#include "action-scripts.h"
 
 #include "asm/dump.h"
 
@@ -95,6 +95,11 @@ bool privately_dump_vma(struct vma_area *vma)
 	if (vma->e->status & VMA_AREA_SYSVIPC)
 		return false;
 
+#ifdef CONFIG_VDSO
+	/* No dumps for vDSO VVAR data */
+	if (vma->e->status & VMA_AREA_VVAR)
+		return false;
+#endif
 	if (vma_area_is(vma, VMA_ANON_SHARED))
 		return false;
 
@@ -129,7 +134,7 @@ void free_mappings(struct vm_area_list *vma_area_list)
 	list_for_each_entry_safe(vma_area, p, &vma_area_list->h, list) {
 		close_vma_file(vma_area);
 		if (!vma_area->file_borrowed)
-			free(vma_area->st);
+			free(vma_area->vmst);
 		free(vma_area);
 	}
 
@@ -205,7 +210,7 @@ static int dump_sched_info(int pid, ThreadCoreEntry *tc)
 	return 0;
 }
 
-struct cr_fdset *glob_fdset;
+struct cr_imgset *glob_imgset;
 
 static int collect_fds(pid_t pid, struct parasite_drain_fd *dfds)
 {
@@ -241,30 +246,43 @@ static int collect_fds(pid_t pid, struct parasite_drain_fd *dfds)
 	return 0;
 }
 
+static int fill_fd_params_special(int fd, struct fd_parms *p)
+{
+	*p = FD_PARMS_INIT;
+
+	if (fstat(fd, &p->stat) < 0) {
+		pr_perror("Can't fstat exe link");
+		return -1;
+	}
+
+	if (get_fd_mntid(fd, &p->mnt_id))
+		return -1;
+
+	return 0;
+}
+
 static int dump_task_exe_link(pid_t pid, MmEntry *mm)
 {
-	struct fd_parms params = FD_PARMS_INIT;
+	struct fd_parms params;
 	int fd, ret = 0;
 
 	fd = open_proc(pid, "exe");
 	if (fd < 0)
 		return -1;
 
-	if (fstat(fd, &params.stat) < 0) {
-		pr_perror("Can't fstat exe link");
+	if (fill_fd_params_special(fd, &params))
 		return -1;
-	}
 
-	if (fd_id_generate_special(&params.stat, &mm->exe_file_id))
+	if (fd_id_generate_special(&params, &mm->exe_file_id))
 		ret = dump_one_reg_file(fd, mm->exe_file_id, &params);
 
 	close(fd);
 	return ret;
 }
 
-static int dump_task_fs(pid_t pid, struct parasite_dump_misc *misc, struct cr_fdset *fdset)
+static int dump_task_fs(pid_t pid, struct parasite_dump_misc *misc, struct cr_imgset *imgset)
 {
-	struct fd_parms p = FD_PARMS_INIT;
+	struct fd_parms p;
 	FsEntry fe = FS_ENTRY__INIT;
 	int fd, ret;
 
@@ -275,12 +293,10 @@ static int dump_task_fs(pid_t pid, struct parasite_dump_misc *misc, struct cr_fd
 	if (fd < 0)
 		return -1;
 
-	if (fstat(fd, &p.stat) < 0) {
-		pr_perror("Can't stat cwd");
+	if (fill_fd_params_special(fd, &p))
 		return -1;
-	}
 
-	if (fd_id_generate_special(&p.stat, &fe.cwd_id)) {
+	if (fd_id_generate_special(&p, &fe.cwd_id)) {
 		ret = dump_one_reg_file(fd, fe.cwd_id, &p);
 		if (ret < 0)
 			return ret;
@@ -292,13 +308,10 @@ static int dump_task_fs(pid_t pid, struct parasite_dump_misc *misc, struct cr_fd
 	if (fd < 0)
 		return -1;
 
-	p = FD_PARMS_INIT;
-	if (fstat(fd, &p.stat) < 0) {
-		pr_perror("Can't stat root");
+	if (fill_fd_params_special(fd, &p))
 		return -1;
-	}
 
-	if (fd_id_generate_special(&p.stat, &fe.root_id)) {
+	if (fd_id_generate_special(&p, &fe.root_id)) {
 		ret = dump_one_reg_file(fd, fe.root_id, &p);
 		if (ret < 0)
 			return ret;
@@ -309,7 +322,7 @@ static int dump_task_fs(pid_t pid, struct parasite_dump_misc *misc, struct cr_fd
 	pr_info("Dumping task cwd id %#x root id %#x\n",
 			fe.cwd_id, fe.root_id);
 
-	return pb_write_one(fdset_fd(fdset, CR_FD_FS), &fe, PB_FS);
+	return pb_write_one(img_from_set(imgset, CR_FD_FS), &fe, PB_FS);
 }
 
 static inline u_int64_t encode_rlim(unsigned long val)
@@ -317,48 +330,56 @@ static inline u_int64_t encode_rlim(unsigned long val)
 	return val == RLIM_INFINITY ? -1 : val;
 }
 
-static int dump_task_rlims(int pid, struct cr_fdset *fds)
+static int dump_task_rlimits(int pid, TaskRlimitsEntry *rls)
 {
-	int res, fd;
+	int res;
 
-	fd = fdset_fd(fds, CR_FD_RLIMIT);
-
-	for (res = 0; res < RLIM_NLIMITS; res++) {
+	for (res = 0; res <rls->n_rlimits ; res++) {
 		struct rlimit lim;
-		RlimitEntry re = RLIMIT_ENTRY__INIT;
 
 		if (prlimit(pid, res, NULL, &lim)) {
 			pr_perror("Can't get rlimit %d", res);
 			return -1;
 		}
 
-		re.cur = encode_rlim(lim.rlim_cur);
-		re.max = encode_rlim(lim.rlim_max);
-
-		if (pb_write_one(fd, &re, PB_RLIMIT))
-			return -1;
+		rls->rlimits[res]->cur = encode_rlim(lim.rlim_cur);
+		rls->rlimits[res]->max = encode_rlim(lim.rlim_max);
 	}
 
 	return 0;
 }
 
 static int dump_filemap(pid_t pid, struct vma_area *vma_area,
-		const struct cr_fdset *fdset)
+		const struct cr_imgset *imgset)
 {
 	struct fd_parms p = FD_PARMS_INIT;
 	VmaEntry *vma = vma_area->e;
 	int ret = 0;
 	u32 id;
 
-	BUG_ON(!vma_area->st);
-	p.stat = *vma_area->st;
+	BUG_ON(!vma_area->vmst);
+	p.stat = *vma_area->vmst;
+	p.mnt_id = vma_area->mnt_id;
 
-	if ((vma->prot & PROT_WRITE) && vma_entry_is(vma, VMA_FILE_SHARED))
-		p.flags = O_RDWR;
-	else
-		p.flags = O_RDONLY;
+	/*
+	 * AUFS support to compensate for the kernel bug
+	 * exposing branch pathnames in map_files.
+	 *
+	 * If the link found in vma_get_mapfile() pointed
+	 * inside a branch, we should use the pathname
+	 * from root that was saved in vma_area->aufs_rpath.
+	 */
+	if (vma_area->aufs_rpath) {
+		struct fd_link aufs_link;
 
-	if (fd_id_generate_special(&p.stat, &id))
+		strcpy(aufs_link.name, vma_area->aufs_rpath);
+		aufs_link.len = strlen(aufs_link.name);
+		p.link = &aufs_link;
+	}
+
+	/* Flags will be set during restore in get_filemap_fd() */
+
+	if (fd_id_generate_special(&p, &id))
 		ret = dump_one_reg_file(vma_area->vm_file_fd, id, &p);
 
 	vma->shmid = id;
@@ -367,7 +388,7 @@ static int dump_filemap(pid_t pid, struct vma_area *vma_area,
 
 static int check_sysvipc_map_dump(pid_t pid, VmaEntry *vma)
 {
-	if (current_ns_mask & CLONE_NEWIPC)
+	if (root_ns_mask & CLONE_NEWIPC)
 		return 0;
 
 	pr_err("Task %d with SysVIPC shmem map @%"PRIx64" doesn't live in IPC ns\n",
@@ -406,7 +427,7 @@ err:
 static int dump_task_mm(pid_t pid, const struct proc_pid_stat *stat,
 		const struct parasite_dump_misc *misc,
 		const struct vm_area_list *vma_area_list,
-		const struct cr_fdset *fdset)
+		const struct cr_imgset *imgset)
 {
 	MmEntry mme = MM_ENTRY__INIT;
 	struct vma_area *vma_area;
@@ -434,7 +455,7 @@ static int dump_task_mm(pid_t pid, const struct proc_pid_stat *stat,
 			ret = add_shmem_area(pid, vma);
 		else if (vma_entry_is(vma, VMA_FILE_PRIVATE) ||
 				vma_entry_is(vma, VMA_FILE_SHARED))
-			ret = dump_filemap(pid, vma_area, fdset);
+			ret = dump_filemap(pid, vma_area, imgset);
 		else if (vma_entry_is(vma, VMA_AREA_SOCKET))
 			ret = dump_socket_map(vma_area);
 		else
@@ -459,6 +480,9 @@ static int dump_task_mm(pid_t pid, const struct proc_pid_stat *stat,
 
 	mme.mm_brk = misc->brk;
 
+	mme.dumpable = misc->dumpable;
+	mme.has_dumpable = true;
+
 	mme.n_mm_saved_auxv = AT_VECTOR_SIZE;
 	mme.mm_saved_auxv = xmalloc(pb_repeated_size(&mme, mm_saved_auxv));
 	if (!mme.mm_saved_auxv)
@@ -470,15 +494,14 @@ static int dump_task_mm(pid_t pid, const struct proc_pid_stat *stat,
 	if (dump_task_exe_link(pid, &mme))
 		goto err;
 
-	ret = pb_write_one(fdset_fd(fdset, CR_FD_MM), &mme, PB_MM);
+	ret = pb_write_one(img_from_set(imgset, CR_FD_MM), &mme, PB_MM);
 	xfree(mme.mm_saved_auxv);
 err:
 	return ret;
 }
 
 static int dump_task_creds(struct parasite_ctl *ctl,
-			   const struct cr_fdset *fds,
-			   struct proc_status_creds *cr)
+			   const struct cr_imgset *fds)
 {
 	CredsEntry ce = CREDS_ENTRY__INIT;
 
@@ -486,30 +509,10 @@ static int dump_task_creds(struct parasite_ctl *ctl,
 	pr_info("Dumping creds for %d)\n", ctl->pid.real);
 	pr_info("----------------------------------------\n");
 
-	ce.uid   = cr->uids[0];
-	ce.gid   = cr->gids[0];
-	ce.euid  = cr->uids[1];
-	ce.egid  = cr->gids[1];
-	ce.suid  = cr->uids[2];
-	ce.sgid  = cr->gids[2];
-	ce.fsuid = cr->uids[3];
-	ce.fsgid = cr->gids[3];
-
-	BUILD_BUG_ON(CR_CAP_SIZE != PROC_CAP_SIZE);
-
-	ce.n_cap_inh = CR_CAP_SIZE;
-	ce.cap_inh = cr->cap_inh;
-	ce.n_cap_prm = CR_CAP_SIZE;
-	ce.cap_prm = cr->cap_prm;
-	ce.n_cap_eff = CR_CAP_SIZE;
-	ce.cap_eff = cr->cap_eff;
-	ce.n_cap_bnd = CR_CAP_SIZE;
-	ce.cap_bnd = cr->cap_bnd;
-
 	if (parasite_dump_creds(ctl, &ce) < 0)
 		return -1;
 
-	return pb_write_one(fdset_fd(fds, CR_FD_CREDS), &ce, PB_CREDS);
+	return pb_write_one(img_from_set(fds, CR_FD_CREDS), &ce, PB_CREDS);
 }
 
 static int get_task_futex_robust_list(pid_t pid, ThreadCoreEntry *info)
@@ -519,39 +522,55 @@ static int get_task_futex_robust_list(pid_t pid, ThreadCoreEntry *info)
 	int ret;
 
 	ret = sys_get_robust_list(pid, &head, &len);
-	if (ret) {
-		pr_err("Failed obtaining futex robust list on %d\n", pid);
-		return -1;
+	if (ret == -ENOSYS) {
+		/*
+		 * If the kernel says get_robust_list is not implemented, then
+		 * check whether set_robust_list is also not implemented, in
+		 * that case we can assume it is empty, since set_robust_list
+		 * is the only way to populate it. This case is possible when
+		 * "futex_cmpxchg_enabled" is unset in the kernel.
+		 *
+		 * The following system call should always fail, even if it is
+		 * implemented, in which case it will return -EINVAL because
+		 * len should be greater than zero.
+		 */
+		if (sys_set_robust_list(NULL, 0) != -ENOSYS)
+			goto err;
+
+		head = NULL;
+		len = 0;
+	} else if (ret) {
+		goto err;
 	}
 
 	info->futex_rla		= encode_pointer(head);
 	info->futex_rla_len	= (u32)len;
 
 	return 0;
+
+err:
+	pr_err("Failed obtaining futex robust list on %d\n", pid);
+	return -1;
 }
 
 static int get_task_personality(pid_t pid, u32 *personality)
 {
-	FILE *file = NULL;
-	int ret = -1;
+	int fd, ret = -1;
 
 	pr_info("Obtaining personality ... ");
 
-	file = fopen_proc(pid, "personality");
-	if (!file)
+	fd = open_proc(pid, "personality");
+	if (fd < 0)
 		goto err;
 
-	if (!fgets(loc_buf, sizeof(loc_buf), file)) {
-		pr_perror("Can't read task personality");
-		goto err;
+	ret = read(fd, loc_buf, sizeof(loc_buf) - 1);
+	close(fd);
+
+	if (ret >= 0) {
+		loc_buf[ret] = '\0';
+		*personality = atoi(loc_buf);
 	}
-
-	*personality = atoi(loc_buf);
-	ret = 0;
-
 err:
-	if (file)
-		fclose(file);
 	return ret;
 }
 
@@ -631,9 +650,9 @@ err:
 	return -1;
 }
 
-static int dump_task_ids(struct pstree_item *item, const struct cr_fdset *cr_fdset)
+static int dump_task_ids(struct pstree_item *item, const struct cr_imgset *cr_imgset)
 {
-	return pb_write_one(fdset_fd(cr_fdset, CR_FD_IDS), item->ids, PB_IDS);
+	return pb_write_one(img_from_set(cr_imgset, CR_FD_IDS), item->ids, PB_IDS);
 }
 
 int dump_thread_core(int pid, CoreEntry *core, const struct parasite_dump_thread *ti)
@@ -649,6 +668,10 @@ int dump_thread_core(int pid, CoreEntry *core, const struct parasite_dump_thread
 		CORE_THREAD_ARCH_INFO(core)->clear_tid_addr = encode_pointer(ti->tid_addr);
 		BUG_ON(!tc->sas);
 		copy_sas(tc->sas, &ti->sas);
+		if (ti->pdeath_sig) {
+			tc->has_pdeath_sig = true;
+			tc->pdeath_sig = ti->pdeath_sig;
+		}
 	}
 
 	return ret;
@@ -657,9 +680,9 @@ int dump_thread_core(int pid, CoreEntry *core, const struct parasite_dump_thread
 static int dump_task_core_all(struct pstree_item *item,
 		const struct proc_pid_stat *stat,
 		const struct parasite_dump_misc *misc,
-		const struct cr_fdset *cr_fdset)
+		const struct cr_imgset *cr_imgset)
 {
-	int fd_core = fdset_fd(cr_fdset, CR_FD_CORE);
+	struct cr_img *img;
 	CoreEntry *core = item->core[0];
 	pid_t pid = item->pid.real;
 	int ret = -1;
@@ -669,7 +692,7 @@ static int dump_task_core_all(struct pstree_item *item,
 	pr_info("----------------------------------------\n");
 
 	ret = get_task_personality(pid, &core->tc->personality);
-	if (ret)
+	if (ret < 0)
 		goto err;
 
 	strncpy((char *)core->tc->comm, stat->comm, TASK_COMM_LEN);
@@ -681,7 +704,17 @@ static int dump_task_core_all(struct pstree_item *item,
 	if (ret)
 		goto err;
 
-	ret = pb_write_one(fd_core, core, PB_CORE);
+	ret = dump_task_rlimits(pid, core->tc->rlimits);
+	if (ret)
+		goto err;
+
+	core->tc->has_cg_set = true;
+	ret = dump_task_cgroup(item, &core->tc->cg_set);
+	if (ret)
+		goto err;
+
+	img = img_from_set(cr_imgset, CR_FD_CORE);
+	ret = pb_write_one(img, core, PB_CORE);
 	if (ret < 0)
 		goto err;
 
@@ -693,10 +726,8 @@ err:
 
 static int parse_children(pid_t pid, pid_t **_c, int *_n)
 {
-	FILE *file;
-	char *tok;
 	pid_t *ch = NULL;
-	int nr = 1;
+	int nr = 0;
 	DIR *dir;
 	struct dirent *de;
 
@@ -705,33 +736,45 @@ static int parse_children(pid_t pid, pid_t **_c, int *_n)
 		return -1;
 
 	while ((de = readdir(dir))) {
+		int fd, len;
+		char *pos;
+
 		if (dir_dots(de))
 			continue;
 
-		file = fopen_proc(pid, "task/%s/children", de->d_name);
-		if (!file)
+		fd = open_proc(pid, "task/%s/children", de->d_name);
+		if (fd < 0)
 			goto err;
 
-		if (!(fgets(loc_buf, sizeof(loc_buf), file)))
-			loc_buf[0] = 0;
+		len = read(fd, loc_buf, sizeof(loc_buf) - 1);
+		close(fd);
+		if (len < 0)
+			goto err;
 
-		fclose(file);
+		loc_buf[len] = '\0';
+		pos = loc_buf;
+		while (1) {
+			pid_t val, *tmp;
 
-		tok = strtok(loc_buf, " \n");
-		while (tok) {
-			pid_t *tmp = xrealloc(ch, nr * sizeof(pid_t));
+			val = strtol(pos, &pos, 0);
+			if (!val) {
+				BUG_ON(*pos != '\0');
+				break;
+			}
+
+			tmp = xrealloc(ch, (nr + 1) * sizeof(pid_t));
 			if (!tmp)
 				goto err;
-			ch = tmp;
-			ch[nr - 1] = atoi(tok);
-			nr++;
-			tok = strtok(NULL, " \n");
-		}
 
+			ch = tmp;
+			ch[nr] = val;
+			nr++;
+			pos++; /* space goes after each pid */
+		}
 	}
 
 	*_c = ch;
-	*_n = nr - 1;
+	*_n = nr;
 
 	closedir(dir);
 	return 0;
@@ -741,12 +784,22 @@ err:
 	return -1;
 }
 
+static inline bool child_collected(struct pstree_item *i, pid_t pid)
+{
+	struct pstree_item *c;
+
+	list_for_each_entry(c, &i->children, sibling)
+		if (c->pid.real == pid)
+			return true;
+
+	return false;
+}
+
 static int collect_task(struct pstree_item *item);
-static int get_children(struct pstree_item *item)
+static int collect_children(struct pstree_item *item)
 {
 	pid_t *ch;
 	int ret, i, nr_children, nr_inprogress;
-	struct pstree_item *c;
 
 	ret = parse_children(item->pid.real, &ch, &nr_children);
 	if (ret < 0)
@@ -754,14 +807,11 @@ static int get_children(struct pstree_item *item)
 
 	nr_inprogress = 0;
 	for (i = 0; i < nr_children; i++) {
+		struct pstree_item *c;
 		pid_t pid = ch[i];
 
 		/* Is it already frozen? */
-		list_for_each_entry(c, &item->children, sibling)
-			if (c->pid.real == pid)
-				break;
-
-		if (&c->sibling != &item->children)
+		if (child_collected(item, pid))
 			continue;
 
 		nr_inprogress++;
@@ -774,7 +824,7 @@ static int get_children(struct pstree_item *item)
 			goto free;
 		}
 
-		ret = seize_task(pid, item->pid.real, &item->pgid, &item->sid);
+		ret = seize_task(pid, item->pid.real);
 		if (ret < 0) {
 			/*
 			 * Here is a race window between parse_children() and seize(),
@@ -788,7 +838,7 @@ static int get_children(struct pstree_item *item)
 			continue;
 		}
 
-		c->pid.real = ch[i];
+		c->pid.real = pid;
 		c->parent = item;
 		c->state = ret;
 		list_add_tail(&c->sibling, &item->children);
@@ -836,10 +886,28 @@ static pid_t item_ppid(const struct pstree_item *item)
 	return item ? item->pid.real : -1;
 }
 
-static int seize_threads(struct pstree_item *item,
-				struct pid *threads, int nr_threads)
+static inline bool thread_collected(struct pstree_item *i, pid_t tid)
 {
-	int i = 0, ret, j, nr_inprogress;
+	int t;
+
+	if (i->pid.real == tid) /* thread leader is collected as task */
+		return true;
+
+	for (t = 0; t < i->nr_threads; t++)
+		if (tid == i->threads[t].real)
+			return true;
+
+	return false;
+}
+
+static int collect_threads(struct pstree_item *item)
+{
+	struct pid *threads = NULL;
+	int nr_threads = 0, i = 0, ret, nr_inprogress, nr_stopped = 0;
+
+	ret = parse_threads(item->pid.real, &threads, &nr_threads);
+	if (ret < 0)
+		goto err;
 
 	if ((item->state == TASK_DEAD) && (nr_threads > 1)) {
 		pr_err("Zombies with threads are not supported\n");
@@ -859,21 +927,16 @@ static int seize_threads(struct pstree_item *item,
 	nr_inprogress = 0;
 	for (i = 0; i < nr_threads; i++) {
 		pid_t pid = threads[i].real;
-		if (item->pid.real == pid)
+
+		if (thread_collected(item, pid))
 			continue;
 
-		for (j = 0; j < item->nr_threads; j++)
-			if (pid == item->threads[j].real)
-				break;
-
-		if (j != item->nr_threads)
-			continue;
 		nr_inprogress++;
 
 		pr_info("\tSeizing %d's %d thread\n",
 				item->pid.real, pid);
 
-		ret = seize_task(pid, item_ppid(item), NULL, NULL);
+		ret = seize_task(pid, item_ppid(item));
 		if (ret < 0) {
 			/*
 			 * Here is a race window between parse_threads() and seize(),
@@ -895,72 +958,64 @@ static int seize_threads(struct pstree_item *item,
 		}
 
 		if (ret == TASK_STOPPED) {
-			pr_err("Stopped threads not supported\n");
-			goto err;
+			nr_stopped++;
 		}
 	}
 
+	if (nr_stopped && nr_stopped != nr_inprogress) {
+		pr_err("Individually stopped threads not supported\n");
+		goto err;
+	}
+
+	xfree(threads);
 	return nr_inprogress;
+
 err:
+	xfree(threads);
 	return -1;
 }
 
-static int collect_threads(struct pstree_item *item)
+static int collect_loop(struct pstree_item *item,
+		int (*collect)(struct pstree_item *))
 {
-	int ret, attempts = NR_ATTEMPTS;
-	struct pid *t;
-	int nr, nr_inprogress;
+	int attempts = NR_ATTEMPTS, nr_inprogress = 1;
+
+	/*
+	 * While we scan the proc and seize the children/threads
+	 * new ones can appear (with clone(CLONE_PARENT) or with
+	 * pthread_create). Thus, after one go, we need to repeat
+	 * the scan-and-freeze again collecting new arrivals. As
+	 * new guys may appear again we do NR_ATTEMPTS passes and
+	 * fail to seize the item if new tasks/threads still
+	 * appear.
+	 */
 
-	nr_inprogress = 1;
 	while (nr_inprogress > 0 && attempts) {
 		attempts--;
-
-		t = NULL;
-		nr = 0;
-
-		ret = parse_threads(item->pid.real, &t, &nr);
-		if (ret < 0)
-			break;
-
-		nr_inprogress = seize_threads(item, t, nr);
-		xfree(t);
-		if (nr_inprogress < 0)
-			break;
-
+		nr_inprogress = collect(item);
 	}
 
-	if (nr_inprogress && attempts)
-		return -1;
+	/*
+	 * We may fail to collect items or run out of attempts.
+	 * In the former case nr_inprogress will be negative, in
+	 * the latter -- positive. Thus it's enough just to check
+	 * for "no more new stuff" and say "we're OK" if so.
+	 */
 
-	return 0;
+	return (nr_inprogress == 0) ? 0 : -1;
 }
 
 static int collect_task(struct pstree_item *item)
 {
-	int ret, nr_inprogress, attempts = NR_ATTEMPTS;
+	int ret;
 
-	ret = collect_threads(item);
+	ret = collect_loop(item, collect_threads);
 	if (ret < 0)
 		goto err_close;
 
-	if (item->state == TASK_DEAD)
-		return 0;
-
 	/* Depth-first search (DFS) is used for traversing a process tree. */
-	nr_inprogress = 1;
-	while (nr_inprogress && attempts) {
-		attempts--;
-
-		/*
-		 * Freeze children and children of children, etc.
-		 * Then check again, that nobody is reparented.
-		 */
-		nr_inprogress = get_children(item);
-		if (nr_inprogress < 0)
-			goto err_close;
-	}
-
-	if (attempts == 0)
+	ret = collect_loop(item, collect_children);
+	if (ret < 0)
 		goto err_close;
 
 	if ((item->state == TASK_DEAD) && !list_empty(&item->children)) {
@@ -968,7 +1023,8 @@ static int collect_task(struct pstree_item *item)
 		goto err_close;
 	}
 
-	close_pid_proc();
+	if (pstree_alloc_cores(item))
+		goto err_close;
 
 	pr_info("Collected %d in %d state\n", item->pid.real, item->state);
 	return 0;
@@ -978,7 +1034,37 @@ err_close:
 	return -1;
 }
 
-static int collect_pstree_ids(void)
+static int collect_pstree_ids_predump(void)
+{
+	struct pstree_item *item;
+	struct {
+		struct pstree_item i;
+		struct dmp_info d;
+	} crt = { };
+
+	/*
+	 * This thing is normally done inside
+	 * write_img_inventory().
+	 */
+
+	crt.i.state = TASK_ALIVE;
+	crt.i.pid.real = getpid();
+
+	if (predump_task_ns_ids(&crt.i))
+		return -1;
+
+	for_each_pstree_item(item) {
+		if (item->state == TASK_DEAD)
+			continue;
+
+		if (predump_task_ns_ids(item))
+			return -1;
+	}
+
+	return 0;
+}
+
+int collect_pstree_ids(void)
 {
 	struct pstree_item *item;
 
@@ -1000,7 +1086,7 @@ static int collect_pstree(pid_t pid)
 		return -1;
 
 	root_item->pid.real = pid;
-	ret = seize_task(pid, -1, &root_item->pgid, &root_item->sid);
+	ret = seize_task(pid, -1);
 	if (ret < 0)
 		goto err;
 	pr_info("Seized task %d, state %d\n", pid, ret);
@@ -1021,36 +1107,7 @@ err:
 
 static int collect_file_locks(void)
 {
-	if (parse_file_locks())
-		return -1;
-
-	if (opts.handle_file_locks)
-		/*
-		 * If the handle file locks option(-l) is set,
-		 * collect work is over.
-		 */
-		return 0;
-
-	/*
-	 * If the handle file locks option is not set, we need to do
-	 * the check, any file locks hold by tasks in our pstree is
-	 * not allowed.
-	 *
-	 * It's hard to do it carefully, there might be some other
-	 * issues like tasks beyond pstree would use flocks hold by
-	 * dumping tasks, but we can't know it in dumping time.
-	 * We need to make sure these flocks only used by dumping tasks.
-	 * We might have to do the check that this option would only
-	 * be used by container dumping.
-	 */
-	if (!list_empty(&file_lock_list)) {
-		pr_err("Some file locks are hold by dumping tasks!"
-			  "You can try --" OPT_FILE_LOCKS " to dump them.\n");
-		return -1;
-	}
-
-	return 0;
-
+	return parse_file_locks();
 }
 
 static int dump_task_thread(struct parasite_ctl *parasite_ctl,
@@ -1059,7 +1116,8 @@ static int dump_task_thread(struct parasite_ctl *parasite_ctl,
 	struct pid *tid = &item->threads[id];
 	CoreEntry *core = item->core[id];
 	pid_t pid = tid->real;
-	int ret = -1, fd_core;
+	int ret = -1;
+	struct cr_img *img;
 
 	pr_info("\n");
 	pr_info("Dumping core for thread (pid: %d)\n", pid);
@@ -1071,13 +1129,13 @@ static int dump_task_thread(struct parasite_ctl *parasite_ctl,
 		goto err;
 	}
 
-	fd_core = open_image(CR_FD_CORE, O_DUMP, tid->virt);
-	if (fd_core < 0)
+	img = open_image(CR_FD_CORE, O_DUMP, tid->virt);
+	if (!img)
 		goto err;
 
-	ret = pb_write_one(fd_core, core, PB_CORE);
+	ret = pb_write_one(img, core, PB_CORE);
 
-	close(fd_core);
+	close_image(img);
 err:
 	pr_info("----------------------------------------\n");
 	return ret;
@@ -1087,7 +1145,8 @@ static int dump_one_zombie(const struct pstree_item *item,
 			   const struct proc_pid_stat *pps)
 {
 	CoreEntry *core;
-	int ret = -1, fd_core;
+	int ret = -1;
+	struct cr_img *img;
 
 	core = core_entry_alloc(0, 1);
 	if (!core)
@@ -1097,91 +1156,116 @@ static int dump_one_zombie(const struct pstree_item *item,
 	core->tc->task_state = TASK_DEAD;
 	core->tc->exit_code = pps->exit_code;
 
-	fd_core = open_image(CR_FD_CORE, O_DUMP, item->pid.virt);
-	if (fd_core < 0)
+	img = open_image(CR_FD_CORE, O_DUMP, item->pid.virt);
+	if (!img)
 		goto err;
 
-	ret = pb_write_one(fd_core, core, PB_CORE);
-	close(fd_core);
+	ret = pb_write_one(img, core, PB_CORE);
+	close_image(img);
 err:
 	core_entry_free(core);
 	return ret;
 }
 
-static int dump_signal_queue(pid_t tid, int fd, bool group)
+#define SI_BATCH	32
+
+static int dump_signal_queue(pid_t tid, SignalQueueEntry **sqe, bool group)
 {
 	struct ptrace_peeksiginfo_args arg;
-	siginfo_t siginfo[32]; /* One page or all non-rt signals */
-	int ret, i = 0, j, nr;
+	int ret;
+	SignalQueueEntry *queue = NULL;
 
 	pr_debug("Dump %s signals of %d\n", group ? "shared" : "private", tid);
 
-	arg.nr = sizeof(siginfo) / sizeof(siginfo_t);
+	arg.nr = SI_BATCH;
 	arg.flags = 0;
 	if (group)
 		arg.flags |= PTRACE_PEEKSIGINFO_SHARED;
+	arg.off = 0;
 
-	for (; ; ) {
-		arg.off = i;
+	queue = xmalloc(sizeof(*queue));
+	if (!queue)
+		return -1;
+
+	signal_queue_entry__init(queue);
+
+	while (1) {
+		int nr, si_pos;
+		siginfo_t *si;
+
+		si = xmalloc(SI_BATCH * sizeof(*si));
+		if (!si) {
+			ret = -1;
+			break;
+		}
+
+		nr = ret = ptrace(PTRACE_PEEKSIGINFO, tid, &arg, si);
+		if (ret == 0)
+			break; /* Finished */
 
-		ret = ptrace(PTRACE_PEEKSIGINFO, tid, &arg, siginfo);
 		if (ret < 0) {
 			if (errno == EIO) {
 				pr_warn("ptrace doesn't support PTRACE_PEEKSIGINFO\n");
 				ret = 0;
 			} else
 				pr_perror("ptrace");
+
 			break;
 		}
 
-		if (ret == 0)
+		queue->n_signals += nr;
+		queue->signals = xrealloc(queue->signals, sizeof(*queue->signals) * queue->n_signals);
+		if (!queue->signals) {
+			ret = -1;
 			break;
-		nr = ret;
+		}
 
-		for (j = 0; j < nr; j++) {
-			SiginfoEntry sie = SIGINFO_ENTRY__INIT;
+		for (si_pos = queue->n_signals - nr;
+				si_pos < queue->n_signals; si_pos++) {
+			SiginfoEntry *se;
 
-			sie.siginfo.len = sizeof(siginfo_t);
-			sie.siginfo.data = (void *) (siginfo + j);
-
-			ret = pb_write_one(fd, &sie, PB_SIGINFO);
-			if (ret < 0)
+			se = xmalloc(sizeof(*se));
+			if (!se) {
+				ret = -1;
 				break;
-			i++;
+			}
+
+			siginfo_entry__init(se);
+			se->siginfo.len = sizeof(siginfo_t);
+			se->siginfo.data = (void *)si++; /* XXX we don't free cores, but when
+							  * we will, this would cause problems
+							  */
+			queue->signals[si_pos] = se;
 		}
+
+		if (ret < 0)
+			break;
+
+		arg.off += nr;
 	}
 
+	*sqe = queue;
 	return ret;
 }
 
-static int dump_thread_signals(struct pid *tid)
-{
-	int fd, ret;
-
-	fd = open_image(CR_FD_PSIGNAL, O_DUMP, tid->virt);
-	if (fd < 0)
-		return -1;
-	ret = dump_signal_queue(tid->real, fd, false);
-	close(fd);
-
-	return ret;
-}
-
-static int dump_task_signals(pid_t pid, struct pstree_item *item,
-		struct cr_fdset *cr_fdset)
+static int dump_task_signals(pid_t pid, struct pstree_item *item)
 {
 	int i, ret;
 
-	ret = dump_signal_queue(pid, fdset_fd(cr_fdset, CR_FD_SIGNAL), true);
-	if (ret) {
-		pr_err("Can't dump pending signals (pid: %d)\n", pid);
-		return -1;
+	/* Dump private signals for each thread */
+	for (i = 0; i < item->nr_threads; i++) {
+		ret = dump_signal_queue(item->threads[i].real, &item->core[i]->thread_core->signals_p, false);
+		if (ret) {
+			pr_err("Can't dump private signals for thread %d\n", item->threads[i].real);
+			return -1;
+		}
 	}
 
-	for (i = 0; i < item->nr_threads; i++) {
-		ret = dump_thread_signals(&item->threads[i]);
-		if (ret)
-			return -1;
+	/* Dump shared signals */
+	ret = dump_signal_queue(pid, &item->core[0]->tc->signals_s, true);
+	if (ret) {
+		pr_err("Can't dump shared signals (pid: %d)\n", pid);
+		return -1;
 	}
 
 	return 0;
@@ -1223,9 +1307,16 @@ static int fill_zombies_pids(struct pstree_item *item)
 	int i, nr;
 	pid_t *ch;
 
+	/*
+	 * Pids read here are virtual -- caller has set up
+	 * the proc of target pid namespace.
+	 */
 	if (parse_children(item->pid.virt, &ch, &nr) < 0)
 		return -1;
 
+	/*
+	 * Step 1 -- filter our ch's pid of alive tasks
+	 */
 	list_for_each_entry(child, &item->children, sibling) {
 		if (child->pid.virt < 0)
 			continue;
@@ -1237,6 +1328,12 @@ static int fill_zombies_pids(struct pstree_item *item)
 		}
 	}
 
+	/*
+	 * Step 2 -- assign remaining pids from ch on
+	 * children's items in arbitrary order. The caller
+	 * will then re-read everything needed to dump
+	 * zombies using newly obtained virtual pids.
+	 */
 	i = 0;
 	list_for_each_entry(child, &item->children, sibling) {
 		if (child->pid.virt > 0)
@@ -1260,7 +1357,7 @@ static int dump_zombies(void)
 {
 	struct pstree_item *item;
 	int ret = -1;
-	int pidns = current_ns_mask & CLONE_NEWPID;
+	int pidns = root_ns_mask & CLONE_NEWPID;
 
 	if (pidns && set_proc_fd(get_service_fd(CR_PROC_FD_OFF)))
 		return -1;
@@ -1335,7 +1432,7 @@ static int pre_dump_one_task(struct pstree_item *item, struct list_head *ctls)
 	}
 
 	ret = -1;
-	parasite_ctl = parasite_infect_seized(pid, item, &vmas, NULL, 0);
+	parasite_ctl = parasite_infect_seized(pid, item, &vmas);
 	if (!parasite_ctl) {
 		pr_err("Can't infect (pid: %d) with parasite\n", pid);
 		goto err_free;
@@ -1384,12 +1481,11 @@ static int dump_one_task(struct pstree_item *item)
 	pid_t pid = item->pid.real;
 	struct vm_area_list vmas;
 	struct parasite_ctl *parasite_ctl;
-	int ret = -1;
+	int ret, exit_code = -1;
 	struct parasite_dump_misc misc;
-	struct cr_fdset *cr_fdset = NULL;
-	struct parasite_drain_fd *dfds;
+	struct cr_imgset *cr_imgset = NULL;
+	struct parasite_drain_fd *dfds = NULL;
 	struct proc_posix_timers_stat proc_args;
-	struct proc_status_creds cr;
 
 	INIT_LIST_HEAD(&vmas.h);
 	vmas.nr = 0;
@@ -1404,35 +1500,29 @@ static int dump_one_task(struct pstree_item *item)
 		 */
 		return 0;
 
-	dfds = xmalloc(sizeof(*dfds));
-	if (!dfds)
-		goto err_free;
-
 	pr_info("Obtaining task stat ... ");
 	ret = parse_pid_stat(pid, &pps_buf);
 	if (ret < 0)
 		goto err;
 
-	ret = parse_pid_status(pid, &cr);
-	if (ret)
-		goto err;
-
-	if (!may_dump(&cr)) {
-		ret = -1;
-		pr_err("Check uid (pid: %d) failed\n", pid);
-		goto err;
-	}
-
 	ret = collect_mappings(pid, &vmas);
 	if (ret) {
 		pr_err("Collect mappings (pid: %d) failed with %d\n", pid, ret);
 		goto err;
 	}
 
-	ret = collect_fds(pid, dfds);
-	if (ret) {
-		pr_err("Collect fds (pid: %d) failed with %d\n", pid, ret);
-		goto err;
+	if (!shared_fdtable(item)) {
+		dfds = xmalloc(sizeof(*dfds));
+		if (!dfds)
+			goto err;
+
+		ret = collect_fds(pid, dfds);
+		if (ret) {
+			pr_err("Collect fds (pid: %d) failed with %d\n", pid, ret);
+			goto err;
+		}
+
+		parasite_ensure_args_size(drain_fds_size(dfds));
 	}
 
 	ret = parse_posix_timers(pid, &proc_args);
@@ -1441,24 +1531,31 @@ static int dump_one_task(struct pstree_item *item)
 		goto err;
 	}
 
-	ret = -1;
-	parasite_ctl = parasite_infect_seized(pid, item, &vmas, dfds, proc_args.timer_n);
+	parasite_ensure_args_size(posix_timers_dump_size(proc_args.timer_n));
+
+	ret = dump_task_signals(pid, item);
+	if (ret) {
+		pr_err("Dump %d signals failed %d\n", pid, ret);
+		goto err;
+	}
+
+	parasite_ctl = parasite_infect_seized(pid, item, &vmas);
 	if (!parasite_ctl) {
 		pr_err("Can't infect (pid: %d) with parasite\n", pid);
 		goto err;
 	}
 
-	if (current_ns_mask & CLONE_NEWPID && root_item == item) {
+	if (root_ns_mask & CLONE_NEWPID && root_item == item) {
 		int pfd;
 
 		pfd = parasite_get_proc_fd_seized(parasite_ctl);
 		if (pfd < 0) {
 			pr_err("Can't get proc fd (pid: %d)\n", pid);
-			goto err_cure_fdset;
+			goto err_cure_imgset;
 		}
 
 		if (install_service_fd(CR_PROC_FD_OFF, pfd) < 0)
-			goto err_cure_fdset;
+			goto err_cure_imgset;
 
 		close(pfd);
 	}
@@ -1466,13 +1563,13 @@ static int dump_one_task(struct pstree_item *item)
 	ret = parasite_fixup_vdso(parasite_ctl, pid, &vmas);
 	if (ret) {
 		pr_err("Can't fixup vdso VMAs (pid: %d)\n", pid);
-		goto err_cure_fdset;
+		goto err_cure_imgset;
 	}
 
 	ret = parasite_dump_misc_seized(parasite_ctl, &misc);
 	if (ret) {
 		pr_err("Can't dump misc (pid: %d)\n", pid);
-		goto err_cure_fdset;
+		goto err_cure_imgset;
 	}
 
 	parasite_ctl->pid.virt = item->pid.virt = misc.pid;
@@ -1485,22 +1582,20 @@ static int dump_one_task(struct pstree_item *item)
 	if (item->sid == 0) {
 		pr_err("A session leader of %d(%d) is outside of its pid namespace\n",
 			item->pid.real, item->pid.virt);
-		ret = -1;
 		goto err_cure;
 	}
 
-	ret = -1;
-	cr_fdset = cr_task_fdset_open(item->pid.virt, O_DUMP);
-	if (!cr_fdset)
+	cr_imgset = cr_task_imgset_open(item->pid.virt, O_DUMP);
+	if (!cr_imgset)
 		goto err_cure;
 
-	ret = dump_task_ids(item, cr_fdset);
+	ret = dump_task_ids(item, cr_imgset);
 	if (ret) {
 		pr_err("Dump ids (pid: %d) failed with %d\n", pid, ret);
 		goto err_cure;
 	}
 
-	if (!shared_fdtable(item)) {
+	if (dfds) {
 		ret = dump_task_files_seized(parasite_ctl, item, dfds);
 		if (ret) {
 			pr_err("Dump files (pid: %d) failed with %d\n", pid, ret);
@@ -1508,52 +1603,49 @@ static int dump_one_task(struct pstree_item *item)
 		}
 	}
 
-	if (opts.handle_file_locks) {
-		ret = dump_task_file_locks(parasite_ctl, cr_fdset, dfds);
-		if (ret) {
-			pr_err("Dump file locks (pid: %d) failed with %d\n",
-				pid, ret);
-			goto err_cure;
-		}
-	}
-
 	ret = parasite_dump_pages_seized(parasite_ctl, &vmas, NULL);
 	if (ret)
 		goto err_cure;
 
-	ret = parasite_dump_sigacts_seized(parasite_ctl, cr_fdset);
+	ret = parasite_dump_sigacts_seized(parasite_ctl, cr_imgset);
 	if (ret) {
 		pr_err("Can't dump sigactions (pid: %d) with parasite\n", pid);
 		goto err_cure;
 	}
 
-	ret = parasite_dump_itimers_seized(parasite_ctl, cr_fdset);
+	ret = parasite_dump_itimers_seized(parasite_ctl, item);
 	if (ret) {
 		pr_err("Can't dump itimers (pid: %d)\n", pid);
 		goto err_cure;
 	}
 
-	ret = parasite_dump_posix_timers_seized(&proc_args, parasite_ctl, cr_fdset);
+	ret = parasite_dump_posix_timers_seized(&proc_args, parasite_ctl, item);
 	if (ret) {
 		pr_err("Can't dump posix timers (pid: %d)\n", pid);
 		goto err_cure;
 	}
 
-	ret = dump_task_core_all(item, &pps_buf, &misc, cr_fdset);
+	ret = dump_task_core_all(item, &pps_buf, &misc, cr_imgset);
 	if (ret) {
 		pr_err("Dump core (pid: %d) failed with %d\n", pid, ret);
 		goto err_cure;
 	}
 
+	ret = dump_task_creds(parasite_ctl, cr_imgset);
+	if (ret) {
+		pr_err("Dump creds (pid: %d) failed with %d\n", pid, ret);
+		goto err;
+	}
+
+	ret = parasite_stop_daemon(parasite_ctl);
+	if (ret) {
+		pr_err("Can't cure (pid: %d) from parasite\n", pid);
+		goto err;
+	}
+
 	ret = dump_task_threads(parasite_ctl, item);
 	if (ret) {
 		pr_err("Can't dump threads\n");
-		goto err_cure;
-	}
-
-	ret = dump_task_creds(parasite_ctl, cr_fdset, &cr);
-	if (ret) {
-		pr_err("Dump creds (pid: %d) failed with %d\n", pid, ret);
 		goto err;
 	}
 
@@ -1563,41 +1655,29 @@ static int dump_one_task(struct pstree_item *item)
 		goto err;
 	}
 
-	ret = dump_task_mm(pid, &pps_buf, &misc, &vmas, cr_fdset);
+	ret = dump_task_mm(pid, &pps_buf, &misc, &vmas, cr_imgset);
 	if (ret) {
 		pr_err("Dump mappings (pid: %d) failed with %d\n", pid, ret);
 		goto err;
 	}
 
-	ret = dump_task_fs(pid, &misc, cr_fdset);
+	ret = dump_task_fs(pid, &misc, cr_imgset);
 	if (ret) {
 		pr_err("Dump fs (pid: %d) failed with %d\n", pid, ret);
 		goto err;
 	}
 
-	ret = dump_task_rlims(pid, cr_fdset);
-	if (ret) {
-		pr_err("Dump %d rlimits failed %d\n", pid, ret);
-		goto err;
-	}
-
-	ret = dump_task_signals(pid, item, cr_fdset);
-	if (ret) {
-		pr_err("Dump %d signals failed %d\n", pid, ret);
-		goto err;
-	}
-
-	close_cr_fdset(&cr_fdset);
+	close_cr_imgset(&cr_imgset);
+	exit_code = 0;
 err:
 	close_pid_proc();
-err_free:
 	free_mappings(&vmas);
 	xfree(dfds);
-	return ret;
+	return exit_code;
 
 err_cure:
-	close_cr_fdset(&cr_fdset);
-err_cure_fdset:
+	close_cr_imgset(&cr_imgset);
+err_cure_imgset:
 	parasite_cure_seized(parasite_ctl);
 	goto err;
 }
@@ -1640,19 +1720,19 @@ int cr_pre_dump_tasks(pid_t pid)
 	if (collect_pstree(pid))
 		goto err;
 
-	if (gen_predump_ns_mask())
+	if (collect_pstree_ids_predump())
 		goto err;
 
-	if (collect_mount_info(pid))
-		goto err;
-
-	if (mntns_collect_root(root_item->pid.real))
+	if (collect_namespaces(false) < 0)
 		goto err;
 
 	for_each_pstree_item(item)
 		if (pre_dump_one_task(item, &ctls))
 			goto err;
 
+	if (irmap_predump_prep())
+		goto err;
+
 	ret = 0;
 err:
 	pstree_switch_state(root_item,
@@ -1691,6 +1771,9 @@ err:
 	if (disconnect_from_page_server())
 		ret = -1;
 
+	if (bfd_flush_images())
+		ret = -1;
+
 	if (ret)
 		pr_err("Pre-dumping FAILED.\n");
 	else {
@@ -1714,7 +1797,7 @@ int cr_dump_tasks(pid_t pid)
 	if (init_stats(DUMP_STATS))
 		goto err;
 
-	if (cr_plugin_init())
+	if (cr_plugin_init(CR_PLUGIN_STAGE__DUMP))
 		goto err;
 
 	if (kerndat_init())
@@ -1729,9 +1812,17 @@ int cr_dump_tasks(pid_t pid)
 	if (vdso_init())
 		goto err;
 
+	if (parse_cg_info())
+		goto err;
+
 	if (write_img_inventory())
 		goto err;
 
+	if (opts.cpu_cap & CPU_CAP_CPU) {
+		if (cpu_dump_cpuinfo())
+			goto err;
+	}
+
 	if (connect_to_page_server())
 		goto err;
 
@@ -1753,17 +1844,11 @@ int cr_dump_tasks(pid_t pid)
 	if (collect_file_locks())
 		goto err;
 
-	if (collect_mount_info(pid))
+	if (collect_namespaces(true) < 0)
 		goto err;
 
-	if (mntns_collect_root(root_item->pid.real))
-		goto err;
-
-	if (collect_sockets(pid))
-		goto err;
-
-	glob_fdset = cr_glob_fdset_open(O_DUMP);
-	if (!glob_fdset)
+	glob_imgset = cr_glob_imgset_open(O_DUMP);
+	if (!glob_imgset)
 		goto err;
 
 	for_each_pstree_item(item) {
@@ -1771,6 +1856,13 @@ int cr_dump_tasks(pid_t pid)
 			goto err;
 	}
 
+	/* MNT namespaces are dumped after files to save remapped links */
+	if (dump_mnt_namespaces() < 0)
+		goto err;
+
+	if (dump_file_locks())
+		goto err;
+
 	if (dump_verify_tty_sids())
 		goto err;
 
@@ -1780,10 +1872,14 @@ int cr_dump_tasks(pid_t pid)
 	if (dump_pstree(root_item))
 		goto err;
 
-	if (current_ns_mask)
-		if (dump_namespaces(root_item, current_ns_mask) < 0)
+	if (root_ns_mask)
+		if (dump_namespaces(root_item, root_ns_mask) < 0)
 			goto err;
 
+	ret = dump_cgroups();
+	if (ret)
+		goto err;
+
 	ret = cr_dump_shmem();
 	if (ret)
 		goto err;
@@ -1801,9 +1897,12 @@ err:
 	if (disconnect_from_page_server())
 		ret = -1;
 
-	close_cr_fdset(&glob_fdset);
+	close_cr_imgset(&glob_imgset);
 
-	cr_plugin_fini();
+	if (bfd_flush_images())
+		ret = -1;
+
+	cr_plugin_fini(CR_PLUGIN_STAGE__DUMP, ret);
 
 	if (!ret) {
 		/*
@@ -1816,7 +1915,7 @@ err:
 		 * Thus ask user via script if we're to break
 		 * checkpoint.
 		 */
-		post_dump_ret = run_scripts("post-dump");
+		post_dump_ret = run_scripts(ACT_POST_DUMP);
 		if (post_dump_ret) {
 			post_dump_ret = WEXITSTATUS(post_dump_ret);
 			pr_info("Post dump script passed with %d\n", post_dump_ret);
@@ -1855,6 +1954,8 @@ err:
 	free_pstree(root_item);
 	free_file_locks();
 	free_link_remaps();
+	free_aufs_branches();
+	free_userns_maps();
 
 	close_service_fd(CR_PROC_FD_OFF);
 
