@@ -3,18 +3,24 @@
 #include <unistd.h>
 
 #include "image.h"
+#include "cr_options.h"
 #include "servicefd.h"
 #include "page-read.h"
 
 #include "protobuf.h"
 #include "protobuf/pagemap.pb-c.h"
 
+#ifndef SEEK_DATA
+#define SEEK_DATA	3
+#define SEEK_HOLE	4
+#endif
+
 static int get_page_vaddr(struct page_read *pr, struct iovec *iov)
 {
 	int ret;
 	u64 img_va;
 
-	ret = read_img_eof(pr->fd_pg, &img_va);
+	ret = read_img_eof(pr->pmi, &img_va);
 	if (ret <= 0)
 		return ret;
 
@@ -28,7 +34,7 @@ static int read_page(struct page_read *pr, unsigned long vaddr, void *buf)
 {
 	int ret;
 
-	ret = read(pr->fd_pg, buf, PAGE_SIZE);
+	ret = read(img_raw_fd(pr->pmi), buf, PAGE_SIZE);
 	if (ret != PAGE_SIZE) {
 		pr_err("Can't read mapping page %d\n", ret);
 		return -1;
@@ -37,18 +43,24 @@ static int read_page(struct page_read *pr, unsigned long vaddr, void *buf)
 	return 1;
 }
 
-static inline void pagemap2iovec(PagemapEntry *pe, struct iovec *iov)
+void pagemap2iovec(PagemapEntry *pe, struct iovec *iov)
 {
 	iov->iov_base = decode_pointer(pe->vaddr);
 	iov->iov_len = pe->nr_pages * PAGE_SIZE;
 }
 
+void iovec2pagemap(struct iovec *iov, PagemapEntry *pe)
+{
+	pe->vaddr = encode_pointer(iov->iov_base);
+	pe->nr_pages = iov->iov_len / PAGE_SIZE;
+}
+
 static int get_pagemap(struct page_read *pr, struct iovec *iov)
 {
 	int ret;
 	PagemapEntry *pe;
 
-	ret = pb_read_one_eof(pr->fd, &pe, PB_PAGEMAP);
+	ret = pb_read_one_eof(pr->pmi, &pe, PB_PAGEMAP);
 	if (ret <= 0)
 		return ret;
 
@@ -79,11 +91,11 @@ static void skip_pagemap_pages(struct page_read *pr, unsigned long len)
 
 	pr_debug("\tpr%u Skip %lx bytes from page-dump\n", pr->id, len);
 	if (!pr->pe->in_parent)
-		lseek(pr->fd_pg, len, SEEK_CUR);
+		lseek(img_raw_fd(pr->pi), len, SEEK_CUR);
 	pr->cvaddr += len;
 }
 
-static int read_pagemap_page_from_parent(struct page_read *pr, unsigned long vaddr, void *buf)
+int seek_pagemap_page(struct page_read *pr, unsigned long vaddr, bool warn)
 {
 	int ret;
 	struct iovec iov;
@@ -96,7 +108,12 @@ static int read_pagemap_page_from_parent(struct page_read *pr, unsigned long vad
 	while (1) {
 		unsigned long iov_end;
 
-		BUG_ON(vaddr < pr->cvaddr);
+		if (vaddr < pr->cvaddr) {
+			if (warn)
+				pr_err("Missing %lu in parent pagemap, current iov: base=%lx,len=%zu\n",
+					vaddr, (unsigned long)iov.iov_base, iov.iov_len);
+			return 0;
+		}
 		iov_end = (unsigned long)iov.iov_base + iov.iov_len;
 
 		if (iov_end <= vaddr) {
@@ -105,13 +122,13 @@ static int read_pagemap_page_from_parent(struct page_read *pr, unsigned long vad
 new_pagemap:
 			ret = get_pagemap(pr, &iov);
 			if (ret <= 0)
-				return -1;
+				return ret;
 
 			continue;
 		}
 
 		skip_pagemap_pages(pr, vaddr - pr->cvaddr);
-		return read_pagemap_page(pr, vaddr, buf);
+		return 1;
 	}
 }
 
@@ -121,15 +138,29 @@ static int read_pagemap_page(struct page_read *pr, unsigned long vaddr, void *bu
 
 	if (pr->pe->in_parent) {
 		pr_debug("\tpr%u Read page %lx from parent\n", pr->id, vaddr);
-		ret = read_pagemap_page_from_parent(pr->parent, vaddr, buf);
+		ret = seek_pagemap_page(pr->parent, vaddr, true);
+		if (ret <= 0)
+			return -1;
+		ret = read_pagemap_page(pr->parent, vaddr, buf);
+		if (ret == -1)
+			return ret;
 	} else {
+		int fd = img_raw_fd(pr->pi);
+		off_t current_vaddr = lseek(fd, 0, SEEK_CUR);
 		pr_debug("\tpr%u Read page %lx from self %lx/%"PRIx64"\n", pr->id,
-				vaddr, pr->cvaddr, lseek(pr->fd_pg, 0, SEEK_CUR));
-		ret = read(pr->fd_pg, buf, PAGE_SIZE);
+				vaddr, pr->cvaddr, current_vaddr);
+		ret = read(fd, buf, PAGE_SIZE);
 		if (ret != PAGE_SIZE) {
 			pr_perror("Can't read mapping page %d", ret);
 			return -1;
 		}
+
+		if (opts.auto_dedup) {
+			ret = punch_hole(pr, current_vaddr, (unsigned int)PAGE_SIZE, false);
+			if (ret == -1) {
+				return -1;
+			}
+		}
 	}
 
 	pr->cvaddr += PAGE_SIZE;
@@ -139,18 +170,27 @@ static int read_pagemap_page(struct page_read *pr, unsigned long vaddr, void *bu
 
 static void close_page_read(struct page_read *pr)
 {
+	int ret;
+
+	if (pr->bunch.iov_len > 0) {
+		ret = punch_hole(pr, 0, 0, true);
+		if (ret == -1)
+			return;
+
+		pr->bunch.iov_len = 0;
+	}
+
 	if (pr->parent) {
 		close_page_read(pr->parent);
 		xfree(pr->parent);
 	}
 
-	close(pr->fd_pg);
-	close(pr->fd);
+	close_image(pr->pmi);
+	if (pr->pi)
+		close_image(pr->pi);
 }
 
-static int open_page_read_at(int dfd, int pid, struct page_read *pr);
-
-static int try_open_parent(int dfd, int pid, struct page_read *pr)
+static int try_open_parent(int dfd, int pid, struct page_read *pr, int flags)
 {
 	int pfd;
 	struct page_read *parent = NULL;
@@ -163,7 +203,7 @@ static int try_open_parent(int dfd, int pid, struct page_read *pr)
 	if (!parent)
 		goto err_cl;
 
-	if (open_page_read_at(pfd, pid, parent)) {
+	if (open_page_read_at(pfd, pid, parent, flags, false)) {
 		if (errno != ENOENT)
 			goto err_free;
 		xfree(parent);
@@ -182,30 +222,33 @@ err_cl:
 	return -1;
 }
 
-static int open_page_read_at(int dfd, int pid, struct page_read *pr)
+int open_page_read_at(int dfd, int pid, struct page_read *pr, int flags, bool shmem)
 {
 	pr->pe = NULL;
+	pr->parent = NULL;
+	pr->bunch.iov_len = 0;
+	pr->bunch.iov_base = NULL;
 
-	pr->fd = open_image_at(dfd, CR_FD_PAGEMAP, O_RSTR, (long)pid);
-	if (pr->fd < 0) {
-		pr->fd_pg = open_image_at(dfd, CR_FD_PAGES_OLD, O_RSTR, pid);
-		if (pr->fd_pg < 0)
+	pr->pmi = open_image_at(dfd, shmem ? CR_FD_SHMEM_PAGEMAP : CR_FD_PAGEMAP, O_RSTR, (long)pid);
+	if (!pr->pmi) {
+		pr->pmi = open_image_at(dfd, shmem ? CR_FD_SHM_PAGES_OLD : CR_FD_PAGES_OLD, flags, pid);
+		if (!pr->pmi)
 			return -1;
 
-		pr->parent = NULL;
 		pr->get_pagemap = get_page_vaddr;
 		pr->put_pagemap = NULL;
 		pr->read_page = read_page;
+		pr->pi = NULL;
 	} else {
 		static unsigned ids = 1;
 
-		if (try_open_parent(dfd, pid, pr)) {
-			close(pr->fd);
+		if (!shmem && try_open_parent(dfd, pid, pr, flags)) {
+			close_image(pr->pmi);
 			return -1;
 		}
 
-		pr->fd_pg = open_pages_image_at(dfd, O_RSTR, pr->fd);
-		if (pr->fd_pg < 0) {
+		pr->pi = open_pages_image_at(dfd, flags, pr->pmi);
+		if (!pr->pi) {
 			close_page_read(pr);
 			return -1;
 		}
@@ -224,7 +267,7 @@ static int open_page_read_at(int dfd, int pid, struct page_read *pr)
 	return 0;
 }
 
-int open_page_read(int pid, struct page_read *pr)
+int open_page_read(int pid, struct page_read *pr, int flags, bool shmem)
 {
-	return open_page_read_at(get_service_fd(IMG_FD_OFF), pid, pr);
+	return open_page_read_at(get_service_fd(IMG_FD_OFF), pid, pr, flags, shmem);
 }
