@@ -2,17 +2,20 @@
 #include <errno.h>
 #include <signal.h>
 #include <linux/limits.h>
+#include <linux/capability.h>
 #include <sys/mount.h>
 #include <stdarg.h>
 #include <sys/ioctl.h>
 
 #include "syscall.h"
 #include "parasite.h"
+#include "config.h"
 #include "fcntl.h"
 #include "prctl.h"
 #include "lock.h"
 #include "vdso.h"
 #include "log.h"
+#include "tty.h"
 
 #include <string.h>
 
@@ -34,6 +37,10 @@ static struct parasite_dump_pages_args *mprotect_args = NULL;
 #define SPLICE_F_GIFT	0x08
 #endif
 
+#ifndef PR_GET_PDEATHSIG
+#define PR_GET_PDEATHSIG  2
+#endif
+
 static int mprotect_vmas(struct parasite_dump_pages_args *args)
 {
 	struct parasite_vma_entry *vmas, *vma;
@@ -142,11 +149,17 @@ static int dump_thread_common(struct parasite_dump_thread *ti)
 {
 	int ret;
 
-	ti->tls = arch_get_tls();
+	arch_get_tls(&ti->tls);
 	ret = sys_prctl(PR_GET_TID_ADDRESS, (unsigned long) &ti->tid_addr, 0, 0, 0);
-	if (ret == 0)
-		ret = sys_sigaltstack(NULL, &ti->sas);
+	if (ret)
+		goto out;
 
+	ret = sys_sigaltstack(NULL, &ti->sas);
+	if (ret)
+		goto out;
+
+	ret = sys_prctl(PR_GET_PDEATHSIG, (unsigned long)&ti->pdeath_sig, 0, 0, 0);
+out:
 	return ret;
 }
 
@@ -159,13 +172,47 @@ static int dump_misc(struct parasite_dump_misc *args)
 	args->pgid = sys_getpgid(0);
 	args->umask = sys_umask(0);
 	sys_umask(args->umask); /* never fails */
+	args->dumpable = sys_prctl(PR_GET_DUMPABLE, 0, 0, 0, 0);
 
 	return dump_thread_common(&args->ti);
 }
 
 static int dump_creds(struct parasite_dump_creds *args)
 {
-	int ret;
+	int ret, i, j;
+	struct cap_data data[_LINUX_CAPABILITY_U32S_3];
+	struct cap_header hdr = {_LINUX_CAPABILITY_VERSION_3, 0};
+
+	ret = sys_capget(&hdr, data);
+	if (ret < 0) {
+		pr_err("Unable to get capabilities: %d\n", ret);
+		return -1;
+	}
+
+	/*
+	 * Loop through the capability constants until we reach cap_last_cap.
+	 * The cap_bnd set is stored as a bitmask comprised of CR_CAP_SIZE number of
+	 * 32-bit uints, hence the inner loop from 0 to 32.
+	 */
+	for (i = 0; i < CR_CAP_SIZE; i++) {
+		args->cap_eff[i] = data[i].eff;
+		args->cap_prm[i] = data[i].prm;
+		args->cap_inh[i] = data[i].inh;
+		args->cap_bnd[i] = 0;
+
+		for (j = 0; j < 32; j++) {
+			if (j + i * 32 > args->cap_last_cap)
+				break;
+			ret = sys_prctl(PR_CAPBSET_READ, j + i * 32, 0, 0, 0);
+			if (ret < 0) {
+				pr_err("Unable to read capability %d: %d\n",
+					j + i * 32, ret);
+				return -1;
+			}
+			if (ret)
+				args->cap_bnd[i] |= (1 << j);
+		}
+	}
 
 	args->secbits = sys_prctl(PR_GET_SECUREBITS, 0, 0, 0, 0);
 
@@ -189,6 +236,22 @@ static int dump_creds(struct parasite_dump_creds *args)
 		return -1;
 	}
 
+	ret = sys_getresuid(&args->uids[0], &args->uids[1], &args->uids[2]);
+	if (ret) {
+		pr_err("Unable to get uids: %d\n", ret);
+		return -1;
+	}
+
+	args->uids[3] = sys_setfsuid(-1L);
+
+	ret = sys_getresgid(&args->gids[0], &args->gids[1], &args->gids[2]);
+	if (ret) {
+		pr_err("Unable to get uids: %d\n", ret);
+		return -1;
+	}
+
+	args->gids[3] = sys_setfsgid(-1L);
+
 	return 0;
 
 grps_err:
@@ -283,60 +346,75 @@ static int parasite_dump_tty(struct parasite_tty_args *args)
 # define TIOCGEXCL	_IOR('T', 0x40, int)
 #endif
 
-	ret = tty_ioctl(args->fd, TIOCGSID, &args->sid);
-	if (ret < 0)
-		goto err;
-
-	ret = tty_ioctl(args->fd, TIOCGPGRP, &args->pgrp);
-	if (ret < 0)
-		goto err;
-
-	ret = tty_ioctl(args->fd, TIOCGPKT, &args->st_pckt);
-	if (ret < 0)
-		goto err;
-
-	ret = tty_ioctl(args->fd, TIOCGPTLCK, &args->st_lock);
-	if (ret < 0)
-		goto err;
-
-	ret = tty_ioctl(args->fd, TIOCGEXCL, &args->st_excl);
-	if (ret < 0)
-		goto err;
-
-	args->hangup = false;
-	return 0;
-
-err:
-	if (ret != -EIO) {
-		pr_err("TTY: Can't get sid/pgrp: %d\n", ret);
-		return -1;
-	}
-
-	/* kernel reports EIO for get ioctls on pair-less ptys */
 	args->sid = 0;
 	args->pgrp = 0;
 	args->st_pckt = 0;
 	args->st_lock = 0;
 	args->st_excl = 0;
-	args->hangup = true;
 
+#define __tty_ioctl(cmd, arg)					\
+	do {							\
+		ret = tty_ioctl(args->fd, cmd, &arg);		\
+		if (ret < 0) {					\
+			if (ret == -ENOTTY)			\
+				arg = 0;			\
+			else if (ret == -EIO)			\
+				goto err_io;			\
+			else					\
+				goto err;			\
+		}						\
+	} while (0)
+
+	__tty_ioctl(TIOCGSID, args->sid);
+	__tty_ioctl(TIOCGPGRP, args->pgrp);
+	__tty_ioctl(TIOCGEXCL,	args->st_excl);
+	
+	if (args->type == TTY_TYPE_PTM ||
+	    args->type == TTY_TYPE_PTS) {
+		__tty_ioctl(TIOCGPKT,	args->st_pckt);
+		__tty_ioctl(TIOCGPTLCK,	args->st_lock);
+	}
+
+	args->hangup = false;
 	return 0;
+
+err:
+	pr_err("tty: Can't fetch params: err = %d\n", ret);
+	return -1;
+err_io:
+
+	/* kernel reports EIO for get ioctls on pair-less ptys */
+	args->hangup = true;
+	return 0;
+#undef __tty_ioctl
 }
 
+#ifdef CONFIG_VDSO
 static int parasite_check_vdso_mark(struct parasite_vdso_vma_entry *args)
 {
 	struct vdso_mark *m = (void *)args->start;
 
 	if (is_vdso_mark(m)) {
+		/*
+		 * Make sure we don't meet some corrupted entry
+		 * where signature matches but verions is not!
+		 */
+		if (m->version != VDSO_MARK_CUR_VERSION) {
+			pr_err("vdso: Mark version mismatch!\n");
+			return -EINVAL;
+		}
 		args->is_marked = 1;
-		args->proxy_addr = m->proxy_addr;
+		args->proxy_vdso_addr = m->proxy_vdso_addr;
+		args->proxy_vvar_addr = m->proxy_vvar_addr;
 	} else {
 		args->is_marked = 0;
-		args->proxy_addr = VDSO_BAD_ADDR;
+		args->proxy_vdso_addr = VDSO_BAD_ADDR;
+		args->proxy_vvar_addr = VVAR_BAD_ADDR;
 	}
 
 	return 0;
 }
+#endif
 
 static int __parasite_daemon_reply_ack(unsigned int cmd, int err)
 {
@@ -380,6 +458,11 @@ static int __parasite_daemon_wait_msg(struct ctl_msg *m)
 	return -1;
 }
 
+static noinline void fini_sigreturn(unsigned long new_sp)
+{
+	ARCH_RT_SIGRETURN(new_sp);
+}
+
 static int fini()
 {
 	unsigned long new_sp;
@@ -396,7 +479,7 @@ static int fini()
 	sys_close(tsock);
 	log_set_fd(-1);
 
-	ARCH_RT_SIGRETURN(new_sp);
+	fini_sigreturn(new_sp);
 
 	BUG();
 
@@ -458,9 +541,11 @@ static noinline __used int noinline parasite_daemon(void *args)
 		case PARASITE_CMD_DUMP_TTY:
 			ret = parasite_dump_tty(args);
 			break;
+#ifdef CONFIG_VDSO
 		case PARASITE_CMD_CHECK_VDSO_MARK:
 			ret = parasite_check_vdso_mark(args);
 			break;
+#endif
 		default:
 			pr_err("Unknown command in parasite daemon thread leader: %d\n", m.cmd);
 			ret = -1;
@@ -501,6 +586,7 @@ static noinline __used int parasite_init_daemon(void *data)
 	struct parasite_init_args *args = data;
 	int ret;
 
+	args->sigreturn_addr = fini_sigreturn;
 	sigframe = args->sigframe;
 
 	tsock = sys_socket(PF_UNIX, SOCK_SEQPACKET, 0);
