@@ -18,6 +18,7 @@
 #include "compiler.h"
 #include "asm/types.h"
 #include "syscall.h"
+#include "config.h"
 #include "prctl.h"
 #include "log.h"
 #include "util.h"
@@ -30,13 +31,18 @@
 #include "restorer.h"
 
 #include "protobuf/creds.pb-c.h"
+#include "protobuf/mm.pb-c.h"
 
 #include "asm/restorer.h"
 
+#ifndef PR_SET_PDEATHSIG
+#define PR_SET_PDEATHSIG 1
+#endif
+
 #define sys_prctl_safe(opcode, val1, val2, val3)			\
 	({								\
 		long __ret = sys_prctl(opcode, val1, val2, val3, 0);	\
-		if (__ret) 						\
+		if (__ret)						\
 			 pr_err("prctl failed @%d with %ld\n", __LINE__, __ret);\
 		__ret;							\
 	})
@@ -45,6 +51,8 @@ static struct task_entries *task_entries;
 static futex_t thread_inprogress;
 static futex_t zombies_inprogress;
 static int cap_last_cap;
+static pid_t *helpers;
+static int n_helpers;
 
 extern void cr_restore_rt (void) asm ("__cr_restore_rt")
 			__attribute__ ((visibility ("hidden")));
@@ -52,6 +60,13 @@ extern void cr_restore_rt (void) asm ("__cr_restore_rt")
 static void sigchld_handler(int signal, siginfo_t *siginfo, void *data)
 {
 	char *r;
+	int i;
+
+	/* We can ignore helpers that die, we expect them to after
+	 * CR_STATE_RESTORE is finished. */
+	for (i = 0; i < n_helpers; i++)
+		if (siginfo->si_pid == helpers[i])
+			return;
 
 	if (futex_get(&task_entries->start) == CR_STATE_RESTORE_SIGCHLD) {
 		pr_debug("%ld: Collect a zombie with (pid %d, %d)\n",
@@ -187,6 +202,60 @@ static int restore_creds(CredsEntry *ce)
 	return 0;
 }
 
+/*
+ * This should be done after creds restore, as
+ * some creds changes might drop the value back
+ * to zero.
+ */
+
+static inline int restore_pdeath_sig(struct thread_restore_args *ta)
+{
+	if (ta->pdeath_sig)
+		return sys_prctl(PR_SET_PDEATHSIG, ta->pdeath_sig, 0, 0, 0);
+	else
+		return 0;
+}
+
+static int restore_dumpable_flag(MmEntry *mme)
+{
+	int current_dumpable;
+	int ret;
+
+	if (!mme->has_dumpable) {
+		pr_warn("Dumpable flag not present in criu dump.\n");
+		return 0;
+	}
+
+	if (mme->dumpable == 0 || mme->dumpable == 1) {
+		ret = sys_prctl(PR_SET_DUMPABLE, mme->dumpable, 0, 0, 0);
+		if (ret) {
+			pr_err("Unable to set PR_SET_DUMPABLE: %d\n", ret);
+			return -1;
+		}
+		return 0;
+	}
+
+	/*
+	 * If dumpable flag is present but it is not 0 or 1, then we can not
+	 * use prctl to set it back.  Try to see if it is already correct
+	 * (which is likely if sysctl fs.suid_dumpable is the same when dump
+	 * and restore are run), in which case there is nothing to do.
+	 * Otherwise, set dumpable to 0 which should be a secure fallback.
+	 */
+	current_dumpable = sys_prctl(PR_GET_DUMPABLE, 0, 0, 0, 0);
+	if (mme->dumpable != current_dumpable) {
+		pr_warn("Dumpable flag [%d] does not match current [%d]. "
+			"Will fallback to setting it to 0 to disable it.\n",
+			mme->dumpable, current_dumpable);
+		ret = sys_prctl(PR_SET_DUMPABLE, 0, 0, 0, 0);
+		if (ret) {
+			pr_err("Unable to set PR_SET_DUMPABLE: %d\n", ret);
+			return -1;
+		}
+	}
+	return 0;
+}
+
 static void restore_sched_info(struct rst_sched_param *p)
 {
 	struct sched_param parm;
@@ -256,11 +325,16 @@ static int restore_thread_common(struct rt_sigframe *sigframe,
 	if (restore_nonsigframe_gpregs(&args->gpregs))
 		return -1;
 
-	restore_tls(args->tls);
+	restore_tls(&args->tls);
 
 	return 0;
 }
 
+static void noinline rst_sigreturn(unsigned long new_sp)
+{
+	ARCH_RT_SIGRETURN(new_sp);
+}
+
 /*
  * Threads restoration via sigreturn. Note it's locked
  * routine and calls for unlock at the end.
@@ -295,6 +369,10 @@ long __export_restore_thread(struct thread_restore_args *args)
 	if (ret)
 		goto core_restore_end;
 
+	ret = restore_dumpable_flag(&args->ta->mm);
+	if (ret)
+		goto core_restore_end;
+
 	pr_info("%ld: Restored\n", sys_gettid());
 
 	restore_finish_stage(CR_STATE_RESTORE);
@@ -303,11 +381,12 @@ long __export_restore_thread(struct thread_restore_args *args)
 		goto core_restore_end;
 
 	restore_finish_stage(CR_STATE_RESTORE_SIGCHLD);
+	restore_pdeath_sig(args);
 	restore_finish_stage(CR_STATE_RESTORE_CREDS);
 	futex_dec_and_wake(&thread_inprogress);
 
 	new_sp = (long)rt_sigframe + SIGFRAME_OFFSET;
-	ARCH_RT_SIGRETURN(new_sp);
+	rst_sigreturn(new_sp);
 
 core_restore_end:
 	pr_err("Restorer abnormal termination for %ld\n", sys_getpid());
@@ -469,6 +548,50 @@ static int vma_remap(unsigned long src, unsigned long dst, unsigned long len)
 	return 0;
 }
 
+static int timerfd_arm(struct task_restore_args *args)
+{
+	int i;
+
+	for (i = 0; i < args->timerfd_n; i++) {
+		struct restore_timerfd *t = &args->timerfd[i];
+		int ret;
+
+		pr_debug("timerfd: arm for fd %d (%d)\n", t->fd, i);
+
+		if (t->settime_flags & TFD_TIMER_ABSTIME) {
+			struct timespec ts = { };
+
+			/*
+			 * We might need to adjust value because the checkpoint
+			 * and restore procedure takes some time itself. Note
+			 * we don't adjust nanoseconds, since the result may
+			 * overflow the limit NSEC_PER_SEC FIXME
+			 */
+			if (sys_clock_gettime(t->clockid, &ts)) {
+				pr_err("Can't get current time");
+				return -1;
+			}
+
+			t->val.it_value.tv_sec += (time_t)ts.tv_sec;
+
+			pr_debug("Ajust id %#x it_value(%llu, %llu) -> it_value(%llu, %llu)\n",
+				 t->id, (unsigned long long)ts.tv_sec,
+				 (unsigned long long)ts.tv_nsec,
+				 (unsigned long long)t->val.it_value.tv_sec,
+				 (unsigned long long)t->val.it_value.tv_nsec);
+		}
+
+		ret  = sys_timerfd_settime(t->fd, t->settime_flags, &t->val, NULL);
+		if (t->ticks)
+			ret |= sys_ioctl(t->fd, TFD_IOC_SET_TICKS, (unsigned long)&t->ticks);
+		if (ret) {
+			pr_err("Can't restore ticks/time for timerfd - %d\n", i);
+			return ret;
+		}
+	}
+	return 0;
+}
+
 static int create_posix_timers(struct task_restore_args *args)
 {
 	int ret, i;
@@ -518,16 +641,23 @@ static void restore_posix_timers(struct task_restore_args *args)
 }
 static void *bootstrap_start;
 static unsigned int bootstrap_len;
-static unsigned long vdso_rt_size;
 
+/*
+ * sys_munmap must not return here. The controll process must
+ * trap us on the exit from sys_munmap.
+ */
+#ifdef CONFIG_VDSO
+static unsigned long vdso_rt_size;
 void __export_unmap(void)
 {
 	sys_munmap(bootstrap_start, bootstrap_len - vdso_rt_size);
-	/*
-	 * sys_munmap must not return here. The controll process must
-	 * trap us on the exit from sys_munmap.
-	 */
 }
+#else
+void __export_unmap(void)
+{
+	sys_munmap(bootstrap_start, bootstrap_len);
+}
+#endif
 
 /*
  * This function unmaps all VMAs, which don't belong to
@@ -586,6 +716,29 @@ static int unmap_old_vmas(void *premmapped_addr, unsigned long premmapped_len,
 	return 0;
 }
 
+static int wait_helpers(struct task_restore_args *task_args)
+{
+	int i;
+
+	for (i = 0; i < task_args->n_helpers; i++) {
+		int status;
+		pid_t pid = task_args->helpers[i];
+
+		/* Check that a helper completed. */
+		if (sys_wait4(pid, &status, 0, NULL) == -1) {
+			/* It has been waited in sigchld_handler */
+			continue;
+		}
+		if (!WIFEXITED(status) || WEXITSTATUS(status)) {
+			pr_err("%d exited with non-zero code (%d,%d)\n", pid,
+				WEXITSTATUS(status), WTERMSIG(status));
+			return -1;
+		}
+	}
+
+	return 0;
+}
+
 /*
  * The main routine to restore task via sigreturn.
  * This one is very special, we never return there
@@ -601,6 +754,7 @@ long __export_restore_task(struct task_restore_args *args)
 	unsigned long va;
 
 	struct rt_sigframe *rt_sigframe;
+	struct prctl_mm_map prctl_map;
 	unsigned long new_sp;
 	k_rtsigset_t to_block;
 	pid_t my_pid = sys_getpid();
@@ -608,9 +762,15 @@ long __export_restore_task(struct task_restore_args *args)
 
 	bootstrap_start = args->bootstrap_start;
 	bootstrap_len	= args->bootstrap_len;
+
+#ifdef CONFIG_VDSO
 	vdso_rt_size	= args->vdso_rt_size;
+#endif
 
 	task_entries = args->task_entries;
+	helpers = args->helpers;
+	n_helpers = args->n_helpers;
+	*args->breakpoint = rst_sigreturn;
 
 	ksigfillset(&act.rt_sa_mask);
 	act.rt_sa_handler = sigchld_handler;
@@ -625,10 +785,10 @@ long __export_restore_task(struct task_restore_args *args)
 
 	pr_info("Switched to the restorer %d\n", my_pid);
 
-	if (vdso_remap("rt-vdso", args->vdso_sym_rt.vma_start,
-		       args->vdso_rt_parked_at,
-		       vdso_vma_size(&args->vdso_sym_rt)))
+#ifdef CONFIG_VDSO
+	if (vdso_do_park(&args->vdso_sym_rt, args->vdso_rt_parked_at, vdso_rt_size))
 		goto core_restore_end;
+#endif
 
 	if (unmap_old_vmas((void *)args->premmapped_addr, args->premmapped_len,
 				bootstrap_start, bootstrap_len))
@@ -653,12 +813,6 @@ long __export_restore_task(struct task_restore_args *args)
 		if (vma_remap(vma_premmaped_start(vma_entry),
 				vma_entry->start, vma_entry_len(vma_entry)))
 			goto core_restore_end;
-
-		if (vma_entry_is(vma_entry, VMA_AREA_VDSO)) {
-			if (vdso_proxify("left dumpee", &args->vdso_sym_rt,
-					 vma_entry, args->vdso_rt_parked_at))
-				goto core_restore_end;
-		}
 	}
 
 	/* Shift private vma-s to the right */
@@ -680,12 +834,6 @@ long __export_restore_task(struct task_restore_args *args)
 		if (vma_remap(vma_premmaped_start(vma_entry),
 				vma_entry->start, vma_entry_len(vma_entry)))
 			goto core_restore_end;
-
-		if (vma_entry_is(vma_entry, VMA_AREA_VDSO)) {
-			if (vdso_proxify("right dumpee", &args->vdso_sym_rt,
-					 vma_entry, args->vdso_rt_parked_at))
-				goto core_restore_end;
-		}
 	}
 
 	/*
@@ -708,6 +856,22 @@ long __export_restore_task(struct task_restore_args *args)
 		}
 	}
 
+#ifdef CONFIG_VDSO
+	/*
+	 * Proxify vDSO.
+	 */
+	for (i = 0; i < args->nr_vmas; i++) {
+		if (vma_entry_is(&args->tgt_vmas[i], VMA_AREA_VDSO) ||
+		    vma_entry_is(&args->tgt_vmas[i], VMA_AREA_VVAR)) {
+			if (vdso_proxify("dumpee", &args->vdso_sym_rt,
+					 args->vdso_rt_parked_at,
+					 i, args->tgt_vmas, args->nr_vmas))
+				goto core_restore_end;
+			break;
+		}
+	}
+#endif
+
 	/*
 	 * Walk though all VMAs again to drop PROT_WRITE
 	 * if it was not there.
@@ -718,16 +882,6 @@ long __export_restore_task(struct task_restore_args *args)
 		if (!(vma_entry_is(vma_entry, VMA_AREA_REGULAR)))
 			continue;
 
-		if (vma_entry_is(vma_entry, VMA_ANON_SHARED)) {
-			struct shmem_info *entry;
-
-			entry = find_shmem(args->shmems, args->nr_shmems,
-						  vma_entry->shmid);
-			if (entry && entry->pid == my_pid &&
-			    entry->start == vma_entry->start)
-				futex_set_and_wake(&entry->lock, 1);
-		}
-
 		if (vma_entry->prot & PROT_WRITE)
 			continue;
 
@@ -768,30 +922,56 @@ long __export_restore_task(struct task_restore_args *args)
 	/*
 	 * Tune up the task fields.
 	 */
-	ret |= sys_prctl_safe(PR_SET_NAME, (long)args->comm, 0, 0);
-
-	ret |= sys_prctl_safe(PR_SET_MM, PR_SET_MM_START_CODE,	(long)args->mm.mm_start_code, 0);
-	ret |= sys_prctl_safe(PR_SET_MM, PR_SET_MM_END_CODE,	(long)args->mm.mm_end_code, 0);
-	ret |= sys_prctl_safe(PR_SET_MM, PR_SET_MM_START_DATA,	(long)args->mm.mm_start_data, 0);
-	ret |= sys_prctl_safe(PR_SET_MM, PR_SET_MM_END_DATA,	(long)args->mm.mm_end_data, 0);
-	ret |= sys_prctl_safe(PR_SET_MM, PR_SET_MM_START_STACK,	(long)args->mm.mm_start_stack, 0);
-	ret |= sys_prctl_safe(PR_SET_MM, PR_SET_MM_START_BRK,	(long)args->mm.mm_start_brk, 0);
-	ret |= sys_prctl_safe(PR_SET_MM, PR_SET_MM_BRK,		(long)args->mm.mm_brk, 0);
-	ret |= sys_prctl_safe(PR_SET_MM, PR_SET_MM_ARG_START,	(long)args->mm.mm_arg_start, 0);
-	ret |= sys_prctl_safe(PR_SET_MM, PR_SET_MM_ARG_END,	(long)args->mm.mm_arg_end, 0);
-	ret |= sys_prctl_safe(PR_SET_MM, PR_SET_MM_ENV_START,	(long)args->mm.mm_env_start, 0);
-	ret |= sys_prctl_safe(PR_SET_MM, PR_SET_MM_ENV_END,	(long)args->mm.mm_env_end, 0);
-	ret |= sys_prctl_safe(PR_SET_MM, PR_SET_MM_AUXV,	(long)args->mm_saved_auxv, args->mm_saved_auxv_size);
+	ret = sys_prctl_safe(PR_SET_NAME, (long)args->comm, 0, 0);
 	if (ret)
 		goto core_restore_end;
 
 	/*
-	 * Because of requirements applied from kernel side
-	 * we need to restore /proc/pid/exe symlink late,
-	 * after old existing VMAs are superseded with
-	 * new ones from image file.
+	 * New kernel interface with @PR_SET_MM_MAP will become
+	 * more widespread once kernel get deployed over the world.
+	 * Thus lets be opportunistic and use new inteface as a try.
 	 */
-	ret = restore_self_exe_late(args);
+	prctl_map = (struct prctl_mm_map) {
+		.start_code	= args->mm.mm_start_code,
+		.end_code	= args->mm.mm_end_code,
+		.start_data	= args->mm.mm_start_data,
+		.end_data	= args->mm.mm_end_data,
+		.start_stack	= args->mm.mm_start_stack,
+		.start_brk	= args->mm.mm_start_brk,
+		.brk		= args->mm.mm_brk,
+		.arg_start	= args->mm.mm_arg_start,
+		.arg_end	= args->mm.mm_arg_end,
+		.env_start	= args->mm.mm_env_start,
+		.env_end	= args->mm.mm_env_end,
+		.auxv		= (void *)args->mm_saved_auxv,
+		.auxv_size	= args->mm_saved_auxv_size,
+		.exe_fd		= args->fd_exe_link,
+	};
+	ret = sys_prctl(PR_SET_MM, PR_SET_MM_MAP, (long)&prctl_map, sizeof(prctl_map), 0);
+	if (ret == -EINVAL) {
+		ret  = sys_prctl_safe(PR_SET_MM, PR_SET_MM_START_CODE,	(long)args->mm.mm_start_code, 0);
+		ret |= sys_prctl_safe(PR_SET_MM, PR_SET_MM_END_CODE,	(long)args->mm.mm_end_code, 0);
+		ret |= sys_prctl_safe(PR_SET_MM, PR_SET_MM_START_DATA,	(long)args->mm.mm_start_data, 0);
+		ret |= sys_prctl_safe(PR_SET_MM, PR_SET_MM_END_DATA,	(long)args->mm.mm_end_data, 0);
+		ret |= sys_prctl_safe(PR_SET_MM, PR_SET_MM_START_STACK,	(long)args->mm.mm_start_stack, 0);
+		ret |= sys_prctl_safe(PR_SET_MM, PR_SET_MM_START_BRK,	(long)args->mm.mm_start_brk, 0);
+		ret |= sys_prctl_safe(PR_SET_MM, PR_SET_MM_BRK,		(long)args->mm.mm_brk, 0);
+		ret |= sys_prctl_safe(PR_SET_MM, PR_SET_MM_ARG_START,	(long)args->mm.mm_arg_start, 0);
+		ret |= sys_prctl_safe(PR_SET_MM, PR_SET_MM_ARG_END,	(long)args->mm.mm_arg_end, 0);
+		ret |= sys_prctl_safe(PR_SET_MM, PR_SET_MM_ENV_START,	(long)args->mm.mm_env_start, 0);
+		ret |= sys_prctl_safe(PR_SET_MM, PR_SET_MM_ENV_END,	(long)args->mm.mm_env_end, 0);
+		ret |= sys_prctl_safe(PR_SET_MM, PR_SET_MM_AUXV,	(long)args->mm_saved_auxv, args->mm_saved_auxv_size);
+
+		/*
+		 * Because of requirements applied from kernel side
+		 * we need to restore /proc/pid/exe symlink late,
+		 * after old existing VMAs are superseded with
+		 * new ones from image file.
+		 */
+		ret |= restore_self_exe_late(args);
+	} else
+		sys_close(args->fd_exe_link);
+
 	if (ret)
 		goto core_restore_end;
 
@@ -848,6 +1028,7 @@ long __export_restore_task(struct task_restore_args *args)
 
 			new_sp = restorer_stack(thread_args + i);
 			last_pid_len = vprint_num(last_pid_buf, sizeof(last_pid_buf), thread_args[i].pid - 1, &s);
+			sys_lseek(fd, 0, SEEK_SET);
 			ret = sys_write(fd, s, last_pid_len);
 			if (ret < 0) {
 				pr_err("Can't set last_pid %ld/%s\n", ret, last_pid_buf);
@@ -882,6 +1063,12 @@ long __export_restore_task(struct task_restore_args *args)
 		goto core_restore_end;
 	}
 
+	ret = timerfd_arm(args);
+	if (ret < 0) {
+		pr_err("Can't restore timerfd %ld\n", ret);
+		goto core_restore_end;
+	}
+
 	pr_info("%ld: Restored\n", sys_getpid());
 
 	futex_set(&zombies_inprogress, args->nr_zombies);
@@ -890,6 +1077,9 @@ long __export_restore_task(struct task_restore_args *args)
 
 	futex_wait_while_gt(&zombies_inprogress, 0);
 
+	if (wait_helpers(args) < 0)
+		goto core_restore_end;
+
 	ksigfillset(&to_block);
 	ret = sys_sigprocmask(SIG_SETMASK, &to_block, NULL, sizeof(k_rtsigset_t));
 	if (ret) {
@@ -911,13 +1101,15 @@ long __export_restore_task(struct task_restore_args *args)
 
 	rst_tcp_socks_all(args);
 
-	/* 
+	/*
 	 * Writing to last-pid is CAP_SYS_ADMIN protected,
 	 * turning off TCP repair is CAP_SYS_NED_ADMIN protected,
 	 * thus restore* creds _after_ all of the above.
 	 */
 
 	ret = restore_creds(&args->creds);
+	ret = ret || restore_dumpable_flag(&args->mm);
+	ret = ret || restore_pdeath_sig(args->t);
 
 	futex_set_and_wake(&thread_inprogress, args->nr_threads);
 
@@ -949,7 +1141,6 @@ long __export_restore_task(struct task_restore_args *args)
 
 	restore_posix_timers(args);
 
-	sys_munmap(args->task_entries, TASK_ENTRIES_SIZE);
 	sys_munmap(args->rst_mem, args->rst_mem_size);
 
 	/*
@@ -962,7 +1153,7 @@ long __export_restore_task(struct task_restore_args *args)
 	 * pure assembly since we don't need any additional
 	 * code insns from gcc.
 	 */
-	ARCH_RT_SIGRETURN(new_sp);
+	rst_sigreturn(new_sp);
 
 core_restore_end:
 	futex_abort_and_wake(&task_entries->nr_in_progress);
