@@ -12,7 +12,7 @@
 #include "asm/types.h"
 #include "libnetlink.h"
 #include "cr_options.h"
-#include "fdset.h"
+#include "imgset.h"
 #include "inet_diag.h"
 #include "files.h"
 #include "image.h"
@@ -291,7 +291,7 @@ static int do_dump_one_inet_fd(int lfd, u32 id, const struct fd_parms *p, int fa
 	if (dump_socket_opts(lfd, &skopts))
 		goto err;
 
-	if (pb_write_one(fdset_fd(glob_fdset, CR_FD_INETSK), &ie, PB_INET_SK))
+	if (pb_write_one(img_from_set(glob_imgset, CR_FD_INETSK), &ie, PB_INET_SK))
 		goto err;
 
 	pr_info("Dumping inet socket at %d\n", p->fd);
@@ -547,13 +547,13 @@ static int open_inet_sk(struct file_desc *d)
 			inet_connect(sk, ii))
 		goto err;
 done:
-	futex_dec(&ii->port->users);
+	futex_dec_and_wake(&ii->port->users);
 
 	if (rst_file_params(sk, ie->fown, ie->flags))
 		goto err;
 
 	if (restore_socket_opts(sk, ie->opts))
-		return -1;
+		goto err;
 
 	return sk;
 
@@ -568,10 +568,10 @@ union sockaddr_inet {
 };
 
 static int restore_sockaddr(union sockaddr_inet *sa,
-		int family, uint32_t pb_port, uint32_t *pb_addr)
+		int family, u32 pb_port, u32 *pb_addr)
 {
-	BUILD_BUG_ON(sizeof(sa->v4.sin_addr.s_addr) > PB_ALEN_INET * sizeof(uint32_t));
-	BUILD_BUG_ON(sizeof(sa->v6.sin6_addr.s6_addr) > PB_ALEN_INET6 * sizeof(uint32_t));
+	BUILD_BUG_ON(sizeof(sa->v4.sin_addr.s_addr) > PB_ALEN_INET * sizeof(u32));
+	BUILD_BUG_ON(sizeof(sa->v6.sin6_addr.s6_addr) > PB_ALEN_INET6 * sizeof(u32));
 
 	memzero(sa, sizeof(*sa));
 
