@@ -7,6 +7,10 @@
 # define PAGE_SIZE 4096
 #endif
 
+#ifndef PR_SET_CHILD_SUBREAPER
+# define PR_SET_CHILD_SUBREAPER 36
+#endif
+
 /* set up test */
 extern void test_ext_init(int argc, char **argv);
 extern void test_init(int argc, char **argv);
@@ -119,4 +123,5 @@ extern int tcp_init_server(int family, int *port);
 extern int tcp_accept_server(int sock);
 extern int tcp_init_client(int family, char *servIP, unsigned short servPort);
 
+extern int get_smaps_bits(unsigned long where, unsigned long *flags, unsigned long *madv);
 #endif /* _VIMITESU_H_ */
